@@ -1,9 +1,8 @@
-/* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB & Sasha
+/* Copyright (C) 2000-2006 MySQL AB & Sasha
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
+   the Free Software Foundation; version 2 of the License.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -23,7 +22,9 @@
 
 int max_binlog_dump_events = 0; // unlimited
 my_bool opt_sporadic_binlog_dump_fail = 0;
+#ifndef DBUG_OFF
 static int binlog_dump_count = 0;
+#endif
 
 /*
     fake_rotate_event() builds a fake (=which does not exist physically in any
@@ -92,8 +93,8 @@ static int send_file(THD *thd)
     The client might be slow loading the data, give him wait_timeout to do
     the job
   */
-  old_timeout = thd->net.read_timeout;
-  thd->net.read_timeout = thd->variables.net_wait_timeout;
+  old_timeout= net->read_timeout;
+  my_net_set_read_timeout(net, thd->variables.net_wait_timeout);
 
   /*
     We need net_flush here because the client will not know it needs to send
@@ -137,7 +138,7 @@ static int send_file(THD *thd)
   error = 0;
 
  err:
-  thd->net.read_timeout = old_timeout;
+  my_net_set_read_timeout(net, old_timeout);
   if (fd >= 0)
     (void) my_close(fd, MYF(0));
   if (errmsg)
@@ -363,7 +364,6 @@ void mysql_binlog_send(THD* thd, char* log_ident, my_off_t pos,
     name=0;					// Find first log
 
   linfo.index_file_offset = 0;
-  thd->current_linfo = &linfo;
 
   if (mysql_bin_log.find_log_pos(&linfo, name, 1))
   {
@@ -372,6 +372,10 @@ void mysql_binlog_send(THD* thd, char* log_ident, my_off_t pos,
     goto err;
   }
 
+  pthread_mutex_lock(&LOCK_thread_count);
+  thd->current_linfo = &linfo;
+  pthread_mutex_unlock(&LOCK_thread_count);
+
   if ((file=open_binlog(&log, log_file_name, &errmsg)) < 0)
   {
     my_errno= ER_MASTER_FATAL_ERROR_READING_BINLOG;
@@ -646,7 +650,7 @@ impossible position";
 
 	if (read_packet)
 	{
-	  thd->proc_info = "Sending binlog event to slave";
+	  thd_proc_info(thd, "Sending binlog event to slave");
 	  if (my_net_write(net, (char*)packet->ptr(), packet->length()) )
 	  {
 	    errmsg = "Failed on my_net_write()";
@@ -684,7 +688,7 @@ impossible position";
       bool loop_breaker = 0;
       /* need this to break out of the for loop from switch */
 
-      thd->proc_info = "Finished reading one binlog; switching to next binlog";
+      thd_proc_info(thd, "Finished reading one binlog; switching to next binlog");
       switch (mysql_bin_log.find_next_log(&linfo, 1)) {
       case LOG_INFO_EOF:
 	loop_breaker = (flags & BINLOG_DUMP_NON_BLOCK);
@@ -730,14 +734,14 @@ end:
   (void)my_close(file, MYF(MY_WME));
 
   send_eof(thd);
-  thd->proc_info = "Waiting to finalize termination";
+  thd_proc_info(thd, "Waiting to finalize termination");
   pthread_mutex_lock(&LOCK_thread_count);
   thd->current_linfo = 0;
   pthread_mutex_unlock(&LOCK_thread_count);
   DBUG_VOID_RETURN;
 
 err:
-  thd->proc_info = "Waiting to finalize termination";
+  thd_proc_info(thd, "Waiting to finalize termination");
   end_io_cache(&log);
   /*
     Exclude  iteration through thread list
@@ -883,13 +887,15 @@ int start_slave(THD* thd , MASTER_INFO* mi,  bool net_report)
 
 int stop_slave(THD* thd, MASTER_INFO* mi, bool net_report )
 {
+  DBUG_ENTER("stop_slave");
+  
   int slave_errno;
   if (!thd)
     thd = current_thd;
 
   if (check_access(thd, SUPER_ACL, any_db,0,0,0,0))
-    return 1;
-  thd->proc_info = "Killing slave";
+    DBUG_RETURN(1);
+  thd_proc_info(thd, "Killing slave");
   int thread_mask;
   lock_slave_threads(mi);
   // Get a mask of _running_ threads
@@ -916,18 +922,18 @@ int stop_slave(THD* thd, MASTER_INFO* mi, bool net_report )
                  ER(ER_SLAVE_WAS_NOT_RUNNING));
   }
   unlock_slave_threads(mi);
-  thd->proc_info = 0;
+  thd_proc_info(thd, 0);
 
   if (slave_errno)
   {
     if (net_report)
       my_message(slave_errno, ER(slave_errno), MYF(0));
-    return 1;
+    DBUG_RETURN(1);
   }
   else if (net_report)
     send_ok(thd);
 
-  return 0;
+  DBUG_RETURN(0);
 }
 
 
@@ -1072,7 +1078,7 @@ bool change_master(THD* thd, MASTER_INFO* mi)
     DBUG_RETURN(TRUE);
   }
 
-  thd->proc_info = "Changing master";
+  thd_proc_info(thd, "Changing master");
   LEX_MASTER_INFO* lex_mi= &thd->lex->mi;
   // TODO: see if needs re-write
   if (init_master_info(mi, master_info_file, relay_log_info_file, 0,
@@ -1197,7 +1203,7 @@ bool change_master(THD* thd, MASTER_INFO* mi)
   if (need_relay_log_purge)
   {
     relay_log_purge= 1;
-    thd->proc_info="Purging old relay logs";
+    thd_proc_info(thd, "Purging old relay logs");
     if (purge_relay_logs(&mi->rli, thd,
 			 0 /* not only reset, but also reinit */,
 			 &errmsg))
@@ -1223,9 +1229,6 @@ bool change_master(THD* thd, MASTER_INFO* mi)
       DBUG_RETURN(TRUE);
     }
   }
-  mi->rli.group_master_log_pos = mi->master_log_pos;
-  DBUG_PRINT("info", ("master_log_pos: %lu", (ulong) mi->master_log_pos));
-
   /*
     Coordinates in rli were spoilt by the 'if (need_relay_log_purge)' block,
     so restore them to good values. If we left them to ''/0, that would work;
@@ -1237,6 +1240,7 @@ bool change_master(THD* thd, MASTER_INFO* mi)
     That's why we always save good coords in rli.
   */
   mi->rli.group_master_log_pos= mi->master_log_pos;
+  DBUG_PRINT("info", ("master_log_pos: %lu", (ulong) mi->master_log_pos));
   strmake(mi->rli.group_master_log_name,mi->master_log_name,
 	  sizeof(mi->rli.group_master_log_name)-1);
 
@@ -1260,7 +1264,7 @@ bool change_master(THD* thd, MASTER_INFO* mi)
   pthread_mutex_unlock(&mi->rli.data_lock);
 
   unlock_slave_threads(mi);
-  thd->proc_info = 0;
+  thd_proc_info(thd, 0);
   send_ok(thd);
   DBUG_RETURN(FALSE);
 }
@@ -1335,7 +1339,6 @@ bool mysql_show_binlog_events(THD* thd)
       name=0;					// Find first log
 
     linfo.index_file_offset = 0;
-    thd->current_linfo = &linfo;
 
     if (mysql_bin_log.find_log_pos(&linfo, name, 1))
     {
@@ -1343,9 +1346,18 @@ bool mysql_show_binlog_events(THD* thd)
       goto err;
     }
 
+    pthread_mutex_lock(&LOCK_thread_count);
+    thd->current_linfo = &linfo;
+    pthread_mutex_unlock(&LOCK_thread_count);
+
     if ((file=open_binlog(&log, linfo.log_file_name, &errmsg)) < 0)
       goto err;
 
+    /*
+      to account binlog event header size
+    */
+    thd->variables.max_allowed_packet += MAX_LOG_EVENT_HEADER;
+
     pthread_mutex_lock(log_lock);
 
     /*
@@ -1356,7 +1368,6 @@ bool mysql_show_binlog_events(THD* thd)
        This code will fail on a mixed relay log (one which has Format_desc then
        Rotate then Format_desc).
     */
-
     ev = Log_event::read_log_event(&log,(pthread_mutex_t*)0,description_event);
     if (ev)
     {
@@ -1417,13 +1428,11 @@ err:
   }
 
   if (errmsg)
-  {
     my_error(ER_ERROR_WHEN_EXECUTING_COMMAND, MYF(0),
              "SHOW BINLOG EVENTS", errmsg);
-    DBUG_RETURN(TRUE);
-  }
+  else
+    send_eof(thd);
 
-  send_eof(thd);
   pthread_mutex_lock(&LOCK_thread_count);
   thd->current_linfo = 0;
   pthread_mutex_unlock(&LOCK_thread_count);
@@ -1549,37 +1558,52 @@ err:
   DBUG_RETURN(TRUE);
 }
 
-
+/**
+   Load data's io cache specific hook to be executed
+   before a chunk of data is being read into the cache's buffer
+   The fuction instantianates and writes into the binlog
+   replication events along LOAD DATA processing.
+   
+   @param file  pointer to io-cache
+   @return 0
+*/
 int log_loaded_block(IO_CACHE* file)
 {
+  DBUG_ENTER("log_loaded_block");
   LOAD_FILE_INFO *lf_info;
-  uint block_len ;
-
-  /* file->request_pos contains position where we started last read */
-  char* buffer = (char*) file->request_pos;
-  if (!(block_len = (char*) file->read_end - (char*) buffer))
-    return 0;
-  lf_info = (LOAD_FILE_INFO*) file->arg;
+  uint block_len;
+  /* buffer contains position where we started last read */
+  char* buffer= (char*) my_b_get_buffer_start(file);
+  uint max_event_size= current_thd->variables.max_allowed_packet;
+  lf_info= (LOAD_FILE_INFO*) file->arg;
   if (lf_info->last_pos_in_file != HA_POS_ERROR &&
-      lf_info->last_pos_in_file >= file->pos_in_file)
-    return 0;
-  lf_info->last_pos_in_file = file->pos_in_file;
-  if (lf_info->wrote_create_file)
+      lf_info->last_pos_in_file >= my_b_get_pos_in_file(file))
+    DBUG_RETURN(0);
+  
+  for (block_len= my_b_get_bytes_in_buffer(file); block_len > 0;
+       buffer += min(block_len, max_event_size),
+       block_len -= min(block_len, max_event_size))
   {
-    Append_block_log_event a(lf_info->thd, lf_info->thd->db, buffer,
-                             block_len, lf_info->log_delayed);
-    mysql_bin_log.write(&a);
+    lf_info->last_pos_in_file= my_b_get_pos_in_file(file);
+    if (lf_info->wrote_create_file)
+    {
+      Append_block_log_event a(lf_info->thd, lf_info->thd->db, buffer,
+                               min(block_len, max_event_size),
+                               lf_info->log_delayed);
+      mysql_bin_log.write(&a);
+    }
+    else
+    {
+      Begin_load_query_log_event b(lf_info->thd, lf_info->thd->db,
+                                   buffer,
+                                   min(block_len, max_event_size),
+                                   lf_info->log_delayed);
+      mysql_bin_log.write(&b);
+      lf_info->wrote_create_file= 1;
+      DBUG_SYNC_POINT("debug_lock.created_file_event",10);
+    }
   }
-  else
-  {
-    Begin_load_query_log_event b(lf_info->thd, lf_info->thd->db,
-                                 buffer, block_len,
-                                 lf_info->log_delayed);
-    mysql_bin_log.write(&b);
-    lf_info->wrote_create_file = 1;
-    DBUG_SYNC_POINT("debug_lock.created_file_event",10);
-  }
-  return 0;
+  DBUG_RETURN(0);
 }
 
 #endif /* HAVE_REPLICATION */
